// Directmode Moire
// Author: SkyBerron

//import femto.mode.HiRes16Color;
import femto.mode.Direct;
import femto.Game;
import femto.State;
import femto.input.Button;
import femto.font.TIC80;
//import static femto.hardware.ST7775.*;
import static femto.hardware.ST7775;
import lutdist;

class Source {
    float xc, yc;
    float r;
    float wa, wr, wph;
    
    int x, y;
    int ph;
    int ilut;
}


class Main extends State {
    static final var screen = new Direct(TIC80.font()); // the screenmode we want to draw with
    
    int counter; // variables are automatically initialized to 0 or null
    long timeLast = 0;
    
    int scene = 0;
    
    // Scanline
    int[] scanline;
    
    // Moire palette
    int[] m_pal;
    
    // Moire sources
    int m_nsrc;
    Source[] m_tsource;

    // Look-up dist
    static final int LUT_YSIZE = 256;
    static final int LUT_YMASK = 255;
    static final int LUT_XSIZE = 512;
    static final int LUT_XMASK = 511;

    public static void main(String[] args){
        Game.run( TIC80.font(), new Main() );
    }
    
    void init(){
        screen.clear(0);
        
        // Scanline
        scanline = new int[220];
        
        // Moire palette
        m_pal = new int[256];
        
        for( int i=0; i < 256; i++ ) {
            m_pal[i] = hsl( ( i * 360.0f ) / 256.0f, 1.0f, .5f );
        }
        
        // Moire sources
        m_nsrc = 3;
        float r0 = 50.0f;
        float r1 = 100.0f;
        m_tsource = new Source[m_nsrc];

        for( int i = 0; i < m_nsrc; i++ ){
            Source src = new Source();
            m_tsource[i] = src;
            float a = ( 2.0f * Math.PI * i ) / m_nsrc;
            src.xc = 110.0f + r0 * Math.cos( a );
            src.yc = 88.0f + r0 * Math.sin( a );
            src.r = r1;
            src.wa = .02f * ( i - .5f );
            src.wr = .01f;
            src.wph = .5f;
        }
        
    }

    void update(){
        int halfw = screen.width()/2;
        int halfh = screen.height()/2;
        String txt = "noname";
        
        if( Button.A.justPressed() ){
            screen.clear(0);
            counter = 0;
        }
        if( Button.B.justPressed() ){
            screen.clear(0);
            scene = ( scene + 1 ) % 4;
        }
        
        if( scene < 4 ) {
            int n = ( scene % 2 ) + 2;
            int m = scene / 2;
            txt = "Moire " + n + " sources";

            for( int i = 0; i < m_nsrc; i++ ){
                Source src = m_tsource[i];
                float a = src.wa * counter;
                float r = src.r * ( .5f + .5f * Math.cos( src.wr * counter ) );
                src.x = (int) ( .5f + src.xc + src.r * Math.cos( a ) );
                src.y = (int) ( .5f + src.yc + src.r * Math.sin( a ) );
                if( src.x < 220 - 256 ) src.x = 220 - 256;
                if( src.x > 255 ) src.x = 255;
                if( src.y < 176 - 256 ) src.y = 176 - 256;
                if( src.y > 255 ) src.y = 255;
                src.ph = (int) ( .5f + src.wph * counter );
            }
            
            pointer ptr = (pointer) lutdist.bin();
            int mask = 16;
            if( m > 0 ) mask = 16+8;

            for( int y = 0; y < 176; y++ ) {
                /*
                for( int x = 0; x < 220; x++ ) {
                    scanline[x] = 0;
                }
                */
                for( int i = 0; i < m_nsrc; i++ ) {
                    Source src = m_tsource[i];
                    int dx = 255 - src.x;
                    int dy = src.y - y;
                    if( dy >= 0 ) {
                        if( dy > 255 ) dy = 255;
                        src.ilut = 512 * dy + dx;
                    } else {
                        if( dy < -255 ) dy = -255;
                        src.ilut = -512 * dy + dx;
                    }
                }

                Source src0 = m_tsource[0];
                int ph0 = src0.ph;
                int ilut0 = src0.ilut;

                Source src1 = m_tsource[1];
                int ph1 = src1.ph;
                int ilut1 = src1.ilut;

                Source src2 = m_tsource[2];
                int ph2 = src2.ph;
                int ilut2 = src2.ilut;

                if( n == 2 ) {
                    for( int x = 0; x < 220; x++ ) {
                        scanline[x] = m_pal[ ( ( ( ph0 + (int) System.memory.LDRB( ptr + ilut0 + x ) ) & mask ) << 3 )
                            ^ ( ( ( ph1 + (int) System.memory.LDRB( ptr + ilut1 + x ) ) & mask ) << 3 )
                            //^ ( ( ( ph2 + (int) System.memory.LDRB( ptr + ilut2 + x ) ) & mask ) << 3 )
                        ];
                    }
                } else {
                    for( int x = 0; x < 220; x++ ) {
                        scanline[x] = m_pal[ ( ( ( ph0 + (int) System.memory.LDRB( ptr + ilut0 + x ) ) & mask ) << 3 )
                            ^ ( ( ( ph1 + (int) System.memory.LDRB( ptr + ilut1 + x ) ) & mask ) << 3 )
                            ^ ( ( ( ph2 + (int) System.memory.LDRB( ptr + ilut2 + x ) ) & mask ) << 3 )
                        ];
                    }
                }
                drawScanline( 0, y, 220, scanline, 0, 220 );
            }
        }

        if( true ) {
            // Print some text
            int y = screen.height() - 6;
            if( scene < 0 ) {
                screen.fillRect( 0, y, screen.width(), 6, 0 );
            }
            screen.setTextPosition( 0, y );
            screen.textColor++;
            screen.println( "Scene " + scene + ": " + txt );
            screen.setTextPosition( screen.width() - 32, y );
            screen.println( (System.currentTimeMillis() - timeLast) + " ms " );
        }

        counter++;
        timeLast = System.currentTimeMillis();
 
    }
    
    // no built-in abs() for floats
    public float abs( float a )   {
        return( ( a >= 0.0f ) ? a : (-a) );
    }
    
    // no built-in % operator for floats
    public float mod( float a, float b ) {
        // using int, casting to long not working
        int d = (int) ( a / b );
        return( a - b * d );
    }
    
    public float mid( float a, float b, float c ) {
        float d = a;
        if( a < b ) {
            if( b < c ) {
                d = b;
            } else if( a < c ) {
                d = c;
            }
        } else if( c < a ) {
            if( b < c ) {
                d = c;
            } else {
                d = b;
            }
        }
        return d;
    }
    
    public int mid( int a, int b, int c ) {
        int d;
        if( a > b ) {
            d = a; a = b; b = d;
        }
        if( b > c ) {
            d = b; b = c; c = d;
        }
        if( a > b ) {
            return a;
        }
        return b;
    }
    
    // no built-in hsl color space conversion
    public int hsl( float h, float s, float l )
    {
        h = mod( h, 360.0f );
        float c = ( 1.0f - abs( 2.0f*l - 1.0f ) ) * s;
        float x = c * ( 1.0f - abs( mod ( h / 60.0f, 2.0f ) - 1.0f ) );
        float m = l - c / 2.0f;
        
        int u = (int) ( .5f + ( c + m ) * 255.0f );
        int v = (int) ( .5f + ( x + m ) * 255.0f );
        
        int r = 0, g = 0, b = 0;
        
        if( h < 180.0f ) {
            if( h < 60.0f ) {
                r = u; g = v;
            } else if( h < 120.0f ) {
                r = v; g = u;
            } else {
                g = u; b = v;
            }
        } else {
            if( h < 240.0f ) {
                g = v; b = u;
            } else if( h < 300.0f ) {
                r = v; b = u;
            } else {
                r = u; b = v;
            }
        }
        
        return (((int)(r>>3))<<11)
                | (((int)(g>>2))<<5)
                | ((int)(b>>3));
    }

    public static void drawScanline( int x, int y, int w, int[] s, int xs, int xf ) {
        if( xs < 0 ) {
            x -= xs;
            w += xs;
            xs = 0;
        }

        if( x < 0 ) {
            xs -= x;
            w += x;
            x = 0;
        }
        
        if( xs + w >= xf ) {
            w = xf - xs;
        }
        
        if( x + w >= 220 )
            w = 220 - x;
        
        if( y < 0 || y >= 176 || x >= 220 || w <= 0 )
            return;

        ST7775.setX(x);
        ST7775.setY(y);
        ST7775.beginStream();
        while( w-- > 0 ) {
            ST7775.writeData( s[xs++] );
        }
        ST7775.setX(0);
        ST7775.setY(0);
        
    }
    
    public void dorun( int[] s, int w, int x0, int x1, int v ) {
        if( x0 < 0 )
            x0 = 0;
        if( x1 >= w )
            x1 = w - 1;
        for( int i = x0; i <= x1; i++ ) {
            s[i] = v;
        }
    }

}
